@@ -152,6 +152,7 @@ def untrained_implicit_model_no_val_data(train_val_implicit_data):
 def models_trained_for_one_step(request,
                                 train_val_implicit_sample_data,
                                 movielens_metadata_df,
+                                user_metadata_df,
                                 movielens_implicit_df,
                                 train_val_implicit_pandas_data,
                                 gpu_count):
@@ -392,7 +393,9 @@ def models_trained_for_one_step(request,
                                              val=val,
                                              item_metadata=movielens_metadata_df,
                                              trained_model=implicit_model,
-                                             metadata_layers_dims=metadata_layers_dims,
+                                             item_metadata_layers_dims=metadata_layers_dims,
+                                             user_metadata=user_metadata_df,
+                                             user_metadata_layers_dims=metadata_layers_dims,
                                              freeze_embeddings=True,
                                              dropout_p=0.15,
                                              loss='warp',
@@ -411,7 +414,9 @@ def models_trained_for_one_step(request,
                                       val=val,
                                       item_metadata=movielens_metadata_df,
                                       trained_model=implicit_model,
-                                      metadata_layers_dims=metadata_layers_dims,
+                                      item_metadata_layers_dims=metadata_layers_dims,
+                                      user_metadata=user_metadata_df,
+                                      user_metadata_layers_dims=metadata_layers_dims,
                                       freeze_embeddings=False,
                                       dropout_p=0.15,
                                       loss='bpr',
@@ -448,7 +453,9 @@ def models_trained_for_one_step(request,
                             val=val,
                             item_metadata=movielens_metadata_df,
                             embedding_dim=10,
-                            metadata_layers_dims=metadata_layers_dims,
+                            item_metadata_layers_dims=metadata_layers_dims,
+                            user_metadata=user_metadata_df,
+                            user_metadata_layers_dims=metadata_layers_dims,
                             lr=1e-1,
                             optimizer='adam',
                             **additional_kwargs)
@@ -516,6 +523,7 @@ def models_trained_for_one_step(request,
 def explicit_models_trained_for_one_step(request,
                                          train_val_explicit_sample_data,
                                          movielens_metadata_df,
+                                         user_metadata_df,
                                          gpu_count):
     train, val = train_val_explicit_sample_data
 
@@ -598,7 +606,9 @@ def explicit_models_trained_for_one_step(request,
                                              val=val,
                                              item_metadata=movielens_metadata_df,
                                              trained_model=implicit_model,
-                                             metadata_layers_dims=None,
+                                             item_metadata_layers_dims=None,
+                                             user_metadata=user_metadata_df,
+                                             user_metadata_layers_dims=None,
                                              freeze_embeddings=True,
                                              dropout_p=0.15,
                                              loss='mae',
@@ -619,7 +629,9 @@ def explicit_models_trained_for_one_step(request,
                                       val=val,
                                       item_metadata=movielens_metadata_df,
                                       trained_model=implicit_model,
-                                      metadata_layers_dims=None,
+                                      item_metadata_layers_dims=None,
+                                      user_metadata=user_metadata_df,
+                                      user_metadata_layers_dims=None,
                                       freeze_embeddings=False,
                                       dropout_p=0.15,
                                       loss='mse',
@@ -636,7 +648,9 @@ def explicit_models_trained_for_one_step(request,
                             val=val,
                             item_metadata=movielens_metadata_df,
                             embedding_dim=10,
-                            metadata_layers_dims=[16, 12],
+                            item_metadata_layers_dims=[16, 12],
+                            user_metadata=user_metadata_df,
+                            user_metadata_layers_dims=[16, 12],
                             lr=1e-1,
                             loss='mae',
                             optimizer='adam')
